"""
Utility functions for tests
"""

from functools import wraps


def retry_test(max_retries):
    """ Retry the decorated test up to `max_retries` times if failure happens. """

    def decorator(test_item):
        @wraps(test_item)
        def new_test_item(self, *args, **kwargs):
            tries = 1
            while tries < max_retries:
                try:
                    r = test_item(self, *args, **kwargs)
                except self.failureException:
                    tries += 1
                else:
                    return r

            # Final retry.
            return test_item(self, *args, **kwargs)

        return new_test_item

    return decorator



# -*- coding: utf-8 -*-
import pickle
import socket
import os
import tempfile
import unittest
from contextlib import suppress
from copy import copy, deepcopy
from itertools import islice
from math import radians
from pathlib import Path

import numpy as np
from crystals import Atom, AtomicStructure, CenteringType, Crystal, Lattice
from crystals.affine import rotation_matrix, transform, translation_matrix
from crystals.crystal import symmetry_expansion, symmetry_reduction
from .utils import retry_test

np.random.seed(23)


def connection_available():
    """ Returns whether or not an internet connection is available """
    with suppress(OSError):
        try:
            socket.create_connection(("www.google.com", 80))
        except:
            return False
        else:
            return True
    return False


class TestSpglibMethods(unittest.TestCase):
    def test_symmetry_graphite(self):
        """ Test that Crystal.symmetry() works correctly for graphite """
        c = Crystal.from_database("C")
        info = c.symmetry(1e-1)

        supposed = {
            "international_number": 194,
            "hall_number": 488,
            "international_symbol": "P6_3/mmc",
            "international_full": "P 6_3/m 2/m 2/c",
            "hall_symbol": "-P 6c 2c",
            "hm_symbol": "P63/mmc",
            "centering": CenteringType("P"),
            "pointgroup": "6/mmm",
        }

        self.assertDictEqual(info, supposed)

    def test_primitive_for_builtins(self):
        """ Test that all built-in crystal have a primitive cell """
        for name in Crystal.builtins:
            with self.subTest(name):
                c = Crystal.from_database(name)
                prim = c.primitive(symprec=0.1)
                self.assertLessEqual(len(prim), len(c))

    def test_primitive_preserves_subclass(self):
        """ Check that Crystal returned by Crystal.primitive() preserve subclass """

        class TestCrystal(Crystal):
            pass

        c = TestCrystal.from_database("C")
        prim = c.primitive()
        self.assertEqual(TestCrystal, type(prim))

    def test_ideal_for_builtins(self):
        """ Test that all built-in crystal have an ideal cell """
        for name in Crystal.builtins:
            with self.subTest(name):
                # This will raise an error if no idealized cell is found
                c = Crystal.from_database(name).ideal()

    def test_ideal_preserves_subclass(self):
        """ Check that Crystal returned by Crystal.ideal() preserve subclass """

        class TestCrystal(Crystal):
            pass

        c = TestCrystal.from_database("C")
        ideal = c.ideal()
        self.assertEqual(TestCrystal, type(ideal))

    def test_symmetry_operations(self):
        """ Test that the symmetry operations output makes sense """
        identity = np.eye(4)

        for name in Crystal.builtins:
            with self.subTest(name):
                c = Crystal.from_database(name)
                symops = c.symmetry_operations()
                self.assertTrue(np.allclose(identity, symops[0]))

    def test_reciprocal_symmetry_operations(self):
        """ Test that the reciprocal symmetry operations output makes sense """
        identity = np.eye(4)

        for name in Crystal.builtins:
            with self.subTest(name):
                c = Crystal.from_database(name)
                symops = c.reciprocal_symmetry_operations()
                self.assertTrue(np.allclose(identity, symops[0]))


class TestCrystalSpecialMethods(unittest.TestCase):
    def test_str_vs_repr(self):
        """ Test that str and repr are workign as expected """
        for name in Crystal.builtins:
            with self.subTest(name):
                c = Crystal.from_database(name)

                # If small crystal, repr and str should be the same
                if len(c) <= 10:
                    self.assertEqual(repr(c), str(c))
                else:
                    self.assertNotEqual(repr(c), str(c))

    def test_equality(self):
        """ Test that __eq__ works as expected """
        c1 = Crystal.from_database("Pu-alpha")
        c2 = deepcopy(c1)
        self.assertEqual(c1, c2)

        c3 = Crystal.from_database("Pu-epsilon")
        self.assertNotEqual(c1, c3)


class TestCrystalConstructors(unittest.TestCase):
    def test_builtins(self):
        """Test that all names in Crystal.builtins build without errors,
        and that Crystal.source is correctly recorded."""
        for name in Crystal.builtins:
            with self.subTest(name):
                c = Crystal.from_database(name)

                self.assertIn(name, c.source)

    def test_builtins_wrong_name(self):
        """ Test that a name not in Crystal.builtins will raise a ValueError """
        with self.assertRaises(ValueError):
            Crystal.from_database("___")

    def test_substructure_preservation(self):
        """ Test that initializing a crystal with substructures preserves the substructures """
        atoms = [Atom("Ag", [0, 0, 0]), Atom("Ag", [1, 1, 1])]
        substructures = [AtomicStructure(atoms=[Atom("U", [0, 0, 0])])]
        c = Crystal(unitcell=atoms + substructures, lattice_vectors=np.eye(3))

        self.assertEqual(len(c), 3)
        self.assertIn(substructures[0], c.substructures)

    @unittest.skipUnless(connection_available(), "Internet connection is required.")
    def test_from_pdb(self):
        """ Test Crystal.from_pdb constructor """
        c = Crystal.from_pdb("1fbb")
        self.assertIn("1fbb", c.source)

    @unittest.skipUnless(connection_available(), "Internet connection is required.")
    def test_from_cod(self):
        """ Test building a Crystal object from the COD """
        # revision = None and latest revision should give the same Crystal
        c = Crystal.from_cod(1521124)
        c2 = Crystal.from_cod(1521124, revision=176429)

        self.assertEqual(c, c2)

    @unittest.skipUnless(connection_available(), "Internet connection is required.")
    def test_from_cod_new_dir(self):
        """ Test that a cache dir is created by Crystal.from_cod """
        with tempfile.TemporaryDirectory() as temp_dir:
            download_dir = Path(temp_dir) / "test_cod"
            self.assertFalse(download_dir.exists())
            c = Crystal.from_cod(1521124, download_dir=download_dir)
            self.assertTrue(download_dir.exists())


class TestSupercell(unittest.TestCase):
    def test_constructors(self):
        """ Test Supercell constructors for varyous 'builtin' structures """
        for name in islice(Crystal.builtins, 20):
            with self.subTest(name):
                s = Crystal.from_database(name).supercell(2, 2, 2)

                self.assertEqual(len(s), 8 * len(s.crystal))


class TestIndexedBy(unittest.TestCase):
    def test_trivial_reindexing(self):
        """ Test re-indexing a crystal by itself """
        c1 = Crystal.from_database("Pu-gamma")
        c2 = c1.indexed_by(c1)

        self.assertEqual(c1, c2)


# # The tests below are very flaky. Due to floating-point arithmetic
# # being non-deterministic, sometimes comparing atoms does not work
# # and the tests below fail.
# # Because of this, we skip tests when in a continuous integration environment
# # See here for details on environment variables:
# #   https://www.appveyor.com/docs/environment-variables/
# @unittest.skipIf(bool(os.environ.get('CI', False)), reason="Flaky tests")
class TestSymmetryReduction(unittest.TestCase):
    def test_trivial(self):
        """Test that the symmetry_reduction function returns the unit cell when
        there is only one possibility."""
        ucell = set([Atom("H", coords=[0, 0, 0])])
        symops = [np.eye(4)]
        asym_cell = symmetry_reduction(ucell, symops)
        self.assertEqual(ucell, asym_cell)

    @retry_test(max_retries=5)
    def test_simple_translation(self):
        """Test that symmetry_reduction works on a unitcell where two atoms are
        linked by a translation"""
        symops = [np.eye(4), translation_matrix([0, 0, 1 / 3])]
        asym_cell = set([Atom("H", coords=[0, 0, 0])])
        unitcell = set(symmetry_expansion(asym_cell, symmetry_operators=symops))

        asym_cell2 = symmetry_reduction(unitcell, symops)
        self.assertEqual(asym_cell, asym_cell2)

    @retry_test(max_retries=5)
    def test_reciprocity_with_symmetry_expansion(self):
        """ Test that symmetry_reduction is reciprocal to symmetry_expansion """
        structures = ["vo2-m1", "Os", "Na"]  # , 'Nb', 'Pd', 'Zn']
        for s in structures:
            with self.subTest("Crystal " + s):
                cryst = Crystal.from_database(s)
                asym_cell = symmetry_reduction(
                    cryst.unitcell, cryst.symmetry_operations()
                )
                ucell = set(symmetry_expansion(asym_cell, cryst.symmetry_operations()))
                self.assertEqual(set(cryst.unitcell), ucell)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
import collections.abc as abc
import pickle
import unittest
from copy import deepcopy
from itertools import islice

import numpy as np

from crystals import Atom, AtomicStructure, Crystal

np.random.seed(23)


class TestAtomicStructure(unittest.TestCase):
    def setUp(self):
        self.substructure = AtomicStructure(atoms=[Atom("U", [0, 0, 0])])
        self.structure = AtomicStructure(
            atoms=[Atom("Ag", [0, 0, 0]), Atom("Ag", [1, 1, 1])],
            substructures=[self.substructure],
        )

    def test_iteration(self):
        """ Test iteration of AtomicStructure yields from orphan atoms and substructure atoms alike """
        elements = [atm.element for atm in self.structure]
        self.assertTrue(len(elements), 3)

    def test_addition_trivial(self):
        """ Test that the addition of two AtomicStructures, one being empty, works as expected """
        addition = self.structure + AtomicStructure()
        self.assertEqual(addition, self.structure)
        self.assertIsNot(addition, self.structure)

    def test_addition_uniqueness(self):
        """ Test that the addition of two AtomicStructures, works as expected regarding unique atoms """
        self.assertEqual(self.structure + self.structure, self.structure)

    def test_addition(self):
        """ Test the addition of two different AtomicStructures works as expected. """
        new_struct = AtomicStructure(
            atoms=[Atom("U", [0, 1, 0])],
            substructures=[
                AtomicStructure(
                    atoms=[Atom("Ag", [0.5, 0, 0]), Atom("Ag", [1, 0.3, 1])]
                )
            ],
        )

        addition = self.structure + new_struct

        self.assertEqual(len(new_struct) + len(self.structure), len(addition))
        self.assertEqual(
            len(new_struct.atoms) + len(self.structure.atoms), len(addition.atoms)
        )
        self.assertEqual(
            len(new_struct.substructures) + len(self.structure.substructures),
            len(addition.substructures),
        )

    def test_addition_subclasses(self):
        """ Test that the addition of two subclass of AtomicStructures is preserved under addition. """

        class NewAtomicStructure(AtomicStructure):
            pass

        addition = NewAtomicStructure() + NewAtomicStructure()
        self.assertIs(type(addition), NewAtomicStructure)

    def test_truthiness(self):
        """ Test that empty AtomicStructures are falsey, and truthy otherwise. """
        empty_structure = AtomicStructure()
        self.assertFalse(empty_structure)

        self.assertTrue(self.structure)

    def test_trivial_transformation(self):
        """ Test that the identity transformation of an AtomicStructure works as expected. """
        transformed = self.structure.transform(np.eye(3))

        # transformed structure should be different, but equal, to original structure
        self.assertIsNot(transformed, self.structure)
        self.assertEqual(transformed, self.structure)

    def test_transformations_inversions(self):
        """ Test that symmetry operations work as expected when inverted. """
        operator = np.random.random(size=(3, 3))
        inv_op = np.linalg.inv(operator)

        transformed1 = self.structure.transform(operator)
        transformed2 = transformed1.transform(inv_op)

        # transformed2 structure should be different, but equal, to original structure
        self.assertIsNot(transformed2, self.structure)
        self.assertEqual(transformed2, self.structure)

    def test_transform_subclass(self):
        """Test that the object returned by the transform() method is the
        same class as the method caller."""

        class NewAtomicStructure(AtomicStructure):
            pass

        structure = NewAtomicStructure(
            atoms=[Atom("Ag", [0, 0, 0]), Atom("Ag", [1, 1, 1])]
        )
        transformed = structure.transform(np.eye(3))

        self.assertIs(type(transformed), type(structure))

    def test_transformations_correctness(self):
        """ Test that AtomicStructure.transform() works as expected. """
        operator = 2 * np.eye(3)
        transformed = self.structure.transform(operator)

        expected_atoms = [atm.transform(operator) for atm in self.structure]

        for atm in expected_atoms:
            self.assertIn(atm, transformed)

    def test_itersorted(self):
        """ Test that AtomicStructure.itersorted() works as expected """
        sorted_from_structure = list(self.structure.itersorted())
        sorted_from_list = list(sorted(self.structure, key=lambda a: a.element))

        self.assertListEqual(sorted_from_structure, sorted_from_list)

    def test_chemical_composition_trivial(self):
        """ Test that AtomicStructure.chemical_composition works as expected """
        expected = {"U": 1 / 3, "Ag": 2 / 3}
        self.assertDictEqual(self.structure.chemical_composition, expected)

    def test_chemical_composition_add_to_unity(self):
        """ Test that AtomicStructure.chemical_composition always adds up to 1 """
        # Faster to create a large atomic structure from a Crystal object
        # Testing for 10 crystal structures only
        for name in islice(Crystal.builtins, 10):
            with self.subTest("Chemical composition: " + name):
                structure = AtomicStructure(atoms=Crystal.from_database(name))
                self.assertAlmostEqual(sum(structure.chemical_composition.values()), 1)

    def test_chemical_formula(self):
        """ Test that AtomicStructure.chemical_formula is working as expected. """
        self.assertEqual(self.structure.chemical_formula, "Ag2 U")

    def test_chemical_formula_hill_notation(self):
        """ Test that the Hill notation, where elements are alphabetically ordered except C and H, which are first. """
        structure = AtomicStructure(
            atoms=[
                Atom("Ag", [0, 1, 0]),
                Atom("C", [0, 0, 0]),
                Atom("H", [0, 1, 0]),
                Atom("U", [1, 1, 1]),
            ]
        )
        self.assertEqual(structure.chemical_formula, "C H Ag U")

    def test_length(self):
        """ Test the __len__ methods """
        self.assertTrue(len(self.structure), 3)

    def test_containership_substructures(self):
        """ Test that containership works on substructure and atoms separately """
        self.assertIn(self.substructure, self.structure)
        self.assertNotIn(self.structure, self.substructure)

    def test_containership_atoms(self):
        """ Test that atom containership testing is working, even in substructures """
        atm = next(iter(self.substructure))
        self.assertIn(atm, self.structure)

    def test_equality(self):
        """ Test that AtomicStructure is equal to itself but not others """
        self.assertEqual(self.structure, self.structure)
        self.assertEqual(self.structure, deepcopy(self.structure))
        self.assertNotEqual(self.structure, self.substructure)

        # Special case: make structures from Crystals
        c1 = Crystal.from_database("vo2-m1")
        c2 = deepcopy(c1)
        s1 = AtomicStructure(atoms=c1)
        s2 = AtomicStructure(atoms=c2.atoms)
        self.assertEqual(s1, s2)

    def test_array(self):
        """ Test AtomicStructure.__array__ """
        arr = np.array(self.structure)
        self.assertSequenceEqual(arr.shape, (len(self.structure), 4))

    def test_picklable(self):
        """Test that Crystal instances can be pickled, and that the unpickled instance
        is identical to the source"""
        pickled = pickle.dumps(self.structure)
        unpickled = pickle.loads(pickled)
        self.assertEqual(self.structure, unpickled)

    def test_abstract_base_classes(self):
        """ Test that AtomicStructure fits with collections.abc module """
        for abstract_base_class in (abc.Hashable, abc.Iterable, abc.Sized):
            self.assertIsInstance(self.structure, abstract_base_class)

    def test_satisfying(self):
        """ Test the AtomicStructure.satisfying method """
        uranium = self.structure.satisfying(lambda a: a.element == "U")
        silver = self.structure.satisfying(lambda a: a.element == "Ag")

        self.assertEqual(len(uranium), 1)
        self.assertEqual(len(silver), 2)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

import tempfile
import unittest
from random import choice, randint

import numpy as np
from pathlib import Path
from crystals import Crystal

try:
    import ase
except ImportError:
    WITH_ASE = False
else:
    WITH_ASE = True


@unittest.skipIf(not WITH_ASE, "ASE not installed or importable")
class TestAseAtoms(unittest.TestCase):
    def setUp(self):
        name = choice(list(Crystal.builtins))
        self.crystal = Crystal.from_database(name)

    def test_construction(self):
        """ Test that ase_atoms returns without error """
        to_ase = self.crystal.to_ase()
        self.assertEqual(len(self.crystal), len(to_ase))

    def test_back_and_forth(self):
        """ Test conversion to and from ase Atoms """
        to_ase = self.crystal.to_ase()
        crystal2 = Crystal.from_ase(to_ase)

        # ase has different handling of coordinates which can lead to
        # rounding beyond 1e-3. Therefore, we cannot compare directly sets
        # self.assertSetEqual(set(self.crystal), set(crystal2))
        self.assertEqual(len(self.crystal), len(crystal2))


class TestCifWriter(unittest.TestCase):
    def test_idempotence(self):
        """ Test that conversion to CIF of a structure loaded from CIF is idempotent. """
        # Testing on all built-in structure assures us that corner cases
        # are taken care of.
        for name in Crystal.builtins:
            with self.subTest(f"CIF idempotence {name}"):

                cryst = Crystal.from_database(name)
                with tempfile.TemporaryDirectory() as temp_dir:
                    f = Path(temp_dir) / "temp.cif"
                    cryst.to_cif(f)
                    cryst2 = Crystal.from_cif(f)
                    self.assertEqual(cryst, cryst2)


if __name__ == "__main__":
    unittest.main()
